
#include "gp2x.h"

#include "allegro.h"
#include <sys/mman.h>
#include <allegro/platform/aintlnx.h>
#include <dirent.h>
#include "crc32.h"

#include <string.h>

#include "vmachine.h"
#include "audio.h"
#include "voice.h"
#include "cpu.h"
#include "vpp.h"
#include "vdc.h"
#include "keyboard.h"

unsigned long                  gp2x_dev;
static volatile unsigned long *gp2x_memregl;
volatile unsigned short       *gp2x_memregs;

int                         mmsp2_blit_base;
static struct MAPPED_MEMORY mmsp2_blit;

static const int gp2x_keyboard[] = { KEY_0, KEY_1, KEY_2, KEY_3, KEY_4, KEY_5, KEY_6,     KEY_7,    KEY_8,     KEY_9,      KEY_PLUS_PAD,
		                             KEY_A, KEY_B, KEY_C, KEY_D, KEY_E, KEY_F, KEY_G,     KEY_H,    KEY_I,     KEY_J,      KEY_MINUS,
				        			 KEY_K, KEY_L, KEY_M, KEY_N, KEY_O, KEY_P, KEY_Q,     KEY_R,    KEY_S,     KEY_T,      KEY_ASTERISK,
							         KEY_U, KEY_V, KEY_W, KEY_X, KEY_Y, KEY_Z, KEY_SPACE, KEY_STOP, KEY_SLASH, KEY_EQUALS, KEY_SLASH_PAD};

static const char gp2x_keyChar[] = { '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', '+',
		                             'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', '-',
					    		     'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'x',
							         'U', 'V', 'W', 'X', 'Y', 'Z', ' ', '.', '?', '=', '/'};

enum  { GP2X_UP=0x1,       GP2X_LEFT=0x4,       GP2X_DOWN=0x10,  GP2X_RIGHT=0x40,
        GP2X_START=1<<8,   GP2X_SELECT=1<<9,    GP2X_L=1<<10,    GP2X_R=1<<11,
        GP2X_A=1<<12,      GP2X_B=1<<13,        GP2X_X=1<<14,    GP2X_Y=1<<15,
        GP2X_VOL_UP=1<<23, GP2X_VOL_DOWN=1<<22, GP2X_PUSH=1<<27 };

enum Bios { G7000, G7400, C52, JOPAC, Unknown };

bool gp2x_left,	  gp2x_right,    gp2x_up,        gp2x_down;
bool gp2x_L,      gp2x_R,        gp2x_A,         gp2x_Y,     gp2x_B, gp2x_X;
bool gp2x_vol_up, gp2x_vol_down, gp2x_vol_both;
bool gp2x_select, gp2x_start;

char gp2x_state[1024];
bool gp2x_load_state = false;

int GP2X_INITIAL_SIZE  = 50;
int GP2X_INCREASE_SIZE = 10;

Bios gp2x_bios          = G7000;
bool gp2x_show_menu     = false;
bool gp2x_show_keyboard = false;

void gp2x_write_to_file(char *txt)
{
	FILE *fn;
	
	fn = fopen("debug.txt","a");
	
	if (!fn) 
		return;

	fputs(txt,    fn);
	fputs("\r\n", fn);

	fclose(fn);
}

void gp2x_sound_volume(int v)
{
	gp2x_mute();

	if (v  < 0)
		v = 0;
	else if (v > 100)
		v = 100;

	app_data.svolume = v;
	app_data.vvolume = v;

	gp2x_display_vol();

	gp2x_unmute();
}

void gp2x_init()
{
	gp2x_show_menu     = false;
	gp2x_show_keyboard = false;

	gp2x_dev     = open("/dev/mem",   O_RDWR); 
	gp2x_memregl = (unsigned long  *)mmap(0, 0x10000, PROT_READ|PROT_WRITE, MAP_SHARED, gp2x_dev, 0xc0000000);
	gp2x_memregs = (unsigned short *)gp2x_memregl;

	{
		//the following is to work around bug in allegro library which causes it to lock up
		//if HW blitter hasn't done anything since GP2X was switched on.

		gp2x_memregs[0x90a >> 1] = 0xffff; // Enable all video and sound devices
  		gp2x_memregs[0x904 >> 1] |= 1<<10; // Enable FASTIO for the hardware blitter

		#define MESGDSTCTRL      0x0000
		#define MESG_BSTBPP      0x60
		#define MESG_DSTBPP_16  (1<<5)
		#define MESG_DSTENB     (1<<6)
		#define MESGDSTADDR      0x0004
		#define MESGSRCCTRL      0x000c
		#define MESG_INVIDEO    (1<<8)
		#define MESGPATCTRL      0x0020
		#define MESGSIZE         0x002c
		#define MESG_HEIGHT      16
		#define MESGCTRL         0x0030
		#define MESG_XDIR_POS    (1<<8)
		#define MESG_YDIR_POS    (1<<9)
		#define MESG_FFCLR       (1<<10)
		#define MESGSTATUS       0x0034
		#define MESG_BUSY        (1<<0)

		#define mmsp2_blit_putl(addr,value) (*(unsigned long volatile *)(mmsp2_blit_base+(addr)) = (value))
		#define mmsp2_blit_getl(addr)       (*(unsigned long volatile *)(mmsp2_blit_base+(addr)))

		mmsp2_blit.base  = 0xe0020000;
		mmsp2_blit.size  = 0x100;
		mmsp2_blit.perms = PROT_READ|PROT_WRITE;
		mmsp2_blit.data  = mmap(0, 0x100, PROT_READ|PROT_WRITE, MAP_SHARED, gp2x_dev, 0xe0020000);
		
		if (mmsp2_blit.data == MAP_FAILED)
			exit(EXIT_FAILURE);

		mmsp2_blit_base = (int)mmsp2_blit.data;

		mmsp2_blit_putl (MESGDSTCTRL, MESG_DSTBPP_16 | MESG_DSTENB);
		mmsp2_blit_putl (MESGDSTADDR, 0x3101000);
		mmsp2_blit_putl (MESGSRCCTRL, MESG_INVIDEO);
		mmsp2_blit_putl (MESGPATCTRL, 0);
		mmsp2_blit_putl (MESGSIZE, (1 << MESG_HEIGHT) | 8);
		mmsp2_blit_putl (MESGCTRL, (MESG_XDIR_POS) | (MESG_YDIR_POS) | 0xaa | MESG_FFCLR);
		mmsp2_blit_putl (MESGSTATUS, 1);
		
		while (mmsp2_blit_getl(MESGSTATUS) & MESG_BUSY);

		munmap((void *)&mmsp2_blit_base, 0x100);
	}
}

void gp2x_deinit()
{
	munmap((void *)gp2x_memregl, 0x10000);
}

void gp2x_init_sound()
{
	app_data.svolume = 50;
	app_data.vvolume = 50;

	init_sound_stream();
	start_voice();
}

void gp2x_mute()
{
	mute_voice();
	mute_audio();
	init_sound_stream(); //has to be here otherwise voices stop working???
}

void gp2x_unmute()
{
	//init_sound_stream(); //done in gp2x_mute()
	update_voice();
}

void gp2x_reset()
{
	init_cpu();
	init_roms();
	init_vpp();
	clearscr();
}

void gp2x_adjust_volume()
{
	while (gp2x_vol_up || gp2x_vol_down)
	{
		if (gp2x_vol_both) 
			gp2x_sound_volume(0);

  		if (gp2x_vol_up && !gp2x_vol_down) 
			gp2x_sound_volume(app_data.svolume+1);

  		if (!gp2x_vol_up && gp2x_vol_down) 
			gp2x_sound_volume(app_data.svolume-1);

		gp2x_joystick_read();
	}
}

void gp2x_joystick_read()
{
	unsigned long value = (gp2x_memregs[0x1198>>1] & 0x00FF);
	unsigned long res;

	if (value == 0xFD) value=0xFA;
	if (value == 0xF7) value=0xEB;
	if (value == 0xDF) value=0xAF;
	if (value == 0x7F) value=0xBE;

	res =~((gp2x_memregs[0x1184>>1] & 0xFF00) | value | (gp2x_memregs[0x1186>>1] << 16));

	gp2x_vol_up   = (res & GP2X_VOL_UP);
	gp2x_vol_down = (res & GP2X_VOL_DOWN);
	gp2x_vol_both = ((res & GP2X_VOL_UP) && (res & GP2X_VOL_DOWN));
	gp2x_start    = (res & GP2X_START);
	gp2x_select   = (res & GP2X_SELECT);
	gp2x_up       = (res & GP2X_UP);
	gp2x_down     = (res & GP2X_DOWN);
	gp2x_left     = (res & GP2X_LEFT);
	gp2x_right    = (res & GP2X_RIGHT);
	gp2x_L        = (res & GP2X_L);
	gp2x_R        = (res & GP2X_R);
	gp2x_A        = (res & GP2X_A);
	gp2x_X        = (res & GP2X_X);
	gp2x_Y        = (res & GP2X_Y);
	gp2x_B        = (res & GP2X_B);
}

void gp2x_flush_keyboard()
{
	int i;
	for (i = 0; i < 44; i++)
		key[gp2x_keyboard[i]] = false;

	key[KEY_ENTER] = false;
	key[KEY_DEL]   = false;
}

void gp2x_flush_key()
{
	gp2x_left     = false;
	gp2x_right    = false;
	gp2x_up       = false;
	gp2x_down     = false;
	gp2x_L        = false;
	gp2x_R        = false;
	gp2x_A        = false;
	gp2x_Y        = false;
	gp2x_B        = false;
	gp2x_X        = false;
	gp2x_vol_up   = false;
	gp2x_vol_down = false;
	gp2x_vol_both = false;
	gp2x_select   = false;
	gp2x_start    = false;
}

bool gp2x_no_command()
{
	if (gp2x_left)     return false;
	if (gp2x_right)    return false;
	if (gp2x_up)       return false;
	if (gp2x_down)     return false;
	if (gp2x_L)        return false;
	if (gp2x_R)        return false;
	if (gp2x_A)        return false;
	if (gp2x_Y)        return false;
	if (gp2x_B)        return false;
	if (gp2x_X)        return false;
	if (gp2x_vol_up)   return false;
	if (gp2x_vol_down) return false;
	if (gp2x_vol_both) return false;
	if (gp2x_select)   return false;
	if (gp2x_start)    return false;

	return true;
}

void gp2x_wait_for_key_release(bool &key)
{
	gp2x_mute();
	do
	{
		gp2x_joystick_read();
	}
	while(key);

	gp2x_unmute();
}
void gp2x_poll()
{
	if (gp2x_show_keyboard)
	{
		gp2x_perform_keyboard();
		return;
	}

	if (gp2x_load_state)
	{
		loadstate(gp2x_state);
		gp2x_load_state = false;
		return;
	}

	if (gp2x_show_menu)
	{
		gp2x_perform_menu();
		return;
	}

	gp2x_flush_keyboard();
	gp2x_joystick_read();

	if (gp2x_vol_up || gp2x_vol_down)
	{
		gp2x_adjust_volume();
		return;
	}

	if (gp2x_start)
	{
		gp2x_wait_for_key_release(gp2x_start);
		gp2x_show_menu = true;
		return;
	}

	if (gp2x_select)
	{
		gp2x_wait_for_key_release(gp2x_select);
		gp2x_load_rom();
		return;
	}

	if (gp2x_Y)
	{
		gp2x_wait_for_key_release(gp2x_Y);
		gp2x_show_keyboard = true;
		app_data.speed = 1000;
	}
	
	key[KEY_0] = gp2x_L;
	key[KEY_1] = gp2x_R;
}

void gp2x_perform_keyboard()
{
	if (!gp2x_show_keyboard)
		return;

	static int limit   = 750;
	static int counter = 0;

	static double row         = 0.;
	static double col         = 0.;
	static bool   transparent = false;
	static double delta = 0.05;

	static const int blue  = 9+32;
	static const int black = 0+32;
	static const int white = 15+32;
	
	static const int nRow       = 5;
	static const int nCol       = 11;
	static const int square     = 16;
	static const int topSpacer  = square/2 - 4;
	static const int leftSpacer = square/2 + 1;
	static const int width      = nCol * square;
	static const int height     = nRow * square;
	static int top              = 240-height; //at bottom
	static int left             = 7;          //at left

	int i, x,  y, selColour;

	int bottom  = top  + height;
	int right   = left + width;

	//if (gp2x_bios == G7400 || gp2x_bios == JOPAC)
	//	limit = 500;
	//else
	//	limit = 750;

	if (counter++ < limit)
		return;
	counter = 0;

	if (!transparent)
		rectfill(bmp, left,   top,    right, bottom,  blue);

	for (i = 0; i < nRow+1; i++)
		hline(bmp, left,   top+(i*square), right, white);

	for (i = 1; i < nCol; i++)
		vline(bmp, left+(i*square), top, bottom-square, white);

	vline(bmp, left+(0     *square), top, bottom, white);
	vline(bmp, left+((nCol)*square), top, bottom, white);

	textout_centre_ex(bmp, font, "L-Ent R-Clr Sel-Move", left+width/2 ,top+(4*square)+topSpacer, white, -1);

	//Draw Current Selection
	x         = left+(int(col)*square); 
	y         = top +(int(row)*square);
	selColour = (transparent) ? blue : black;
	hline(bmp, x,        y,        x+square, selColour);
	hline(bmp, x,        y+square, x+square, selColour);
	vline(bmp, x,        y,        y+square, selColour);
	vline(bmp, x+square, y,        y+square, selColour);

	char txt[] = {0,0};

	for (i = 0; i < 44; i++)
	{
		txt[0] = gp2x_keyChar[i];
		y      = i / nCol;
		x      = i % nCol;
		textout_centre_ex(bmp, font, txt, left+(x*square)+leftSpacer ,top+(y*square)+topSpacer, white, -1);
	}

	finish_display();

	gp2x_flush_keyboard();
	gp2x_joystick_read();

	if (gp2x_no_command())
		return;

	if (gp2x_start)
	{
		gp2x_wait_for_key_release(gp2x_start);
		transparent = !transparent;
	}

	if (gp2x_select)
	{
		if (gp2x_up)
			--top;
		if (gp2x_down)
			++top;
		if (gp2x_left)
			--left;
		if (gp2x_right)
			++left;

		if (top < 2)
			top = 2;
		if (top > 240-height)
			top = 240-height;
		if (left < 7)
			left = 7;
		if (left > 320-width)
			left = 320-width;
	}
	else
	{
		if (gp2x_up)
			row = row - delta;
		if (gp2x_down)
			row = row + delta;
		if (gp2x_left)
			col = col - delta;
		if (gp2x_right)
			col = col + delta;

		if (row < 0.)
			row = 0.;
		if (row > nRow-2.)
			row = nRow-2.;
		if (col < 0.)
			col = 0.;
		if (col > nCol-1.)
			col = nCol-1.;
	}

	//key[KEY_Y]     = gp2x_A;
	//key[KEY_N]     = gp2x_X;
	key[KEY_ENTER] = gp2x_L;
	key[KEY_DEL]   = gp2x_R;

	if (gp2x_X)
		key[gp2x_keyboard[(int(row)*nCol) + int(col)]] = true;

	if (gp2x_Y)
	{
		gp2x_wait_for_key_release(gp2x_Y);
		gp2x_show_keyboard = false;
		app_data.speed = 100;
	}
}

void gp2x_perform_menu()
{
	gp2x_show_menu = false;

	switch(gp2x_do_menu())
	{
	case 0:
		//reset emulator (same as pressing original consoles reset button)
		gp2x_reset();
		break;
	case 1:
		//exit emulator
		key_done = true;
		break;

	case 2:
		gp2x_loadstate();
		break;

	case 3:
		gp2x_savestate();
		break;

	case 4:
		gp2x_toggle_stick();
		break;

	case 5:
		gp2x_toggle_region();
		break;
	}
}

void gp2x_menu_txt(int x, int y, int fore, const char *msg)
{
	textout_centre_ex(bmp, font, msg, x+0 , y+0, fore, -1);
}

int gp2x_do_menu()
{
	static const int blue  = 9+32;
	static const int white = 0+32;
	static const int black = 15+32;

	int ret    = 5;
	int col    = white;
	int nItems = 6;
	int limit  = 5;

	int top     = 72;
	int textTop = top + 4;
	int left    = 60;
	int right   = left + 200;
	int bottom  = textTop + (14*(nItems+1)) + 2;
	int index   = 0;

	gp2x_mute();

	rectfill(bmp, left,   top,    right, bottom,  blue);
	hline(   bmp, left,   top,    right,          15+32);
	hline(   bmp, left+1, bottom, right,          1+32);

	vline(   bmp, left,   top,    bottom,         15+32);
	vline(   bmp, right,  bottom, top,            1+32);

	do
	{
		index   = 0;
		textTop = top + 4;

		col = (ret == index ? white : black);
		gp2x_menu_txt(160,  textTop, col, "Reset Emulator");
		textTop += 14;
		index++;

		col = (ret == index ? white : black);
		gp2x_menu_txt(160,  textTop, col, "Exit Emulator");
		textTop += 14;
		index++;

		col = (ret == index ? white : black);
		gp2x_menu_txt(160, textTop, col, "Load State");
		textTop += 14;
		index++;

		col = (ret == index ? white : black);
		gp2x_menu_txt(160, textTop, col, "Save State");
		textTop += 14;
		index++;

		col = (ret == index ? white : black);
		gp2x_menu_txt(160, textTop, col, "Switch Sticks");
		textTop += 14;
		index++;

		//col = (ret == index ? white : black);
		//if (app_data.euro == 0)
		//	gp2x_menu_txt(160, textTop, col, "Switch to PAL");
		//else
		//	gp2x_menu_txt(160, textTop, col, "Switch to NTSC");
		//textTop += 14;
		//index++;

		textTop += 14;
		col = (ret == index ? white : black);
		gp2x_menu_txt(160, textTop, col, "Exit Menu");

		finish_display();

		rest(100);
		gp2x_joystick_read();

		if (gp2x_up)
			ret--;

		if (gp2x_down)
			ret++;

		if (ret < 0)
			ret = limit;

		if (ret > limit)
			ret = 0;
	} 
	while(!(gp2x_B || gp2x_start));

	if (gp2x_start)
	{
		gp2x_wait_for_key_release(gp2x_start);
		ret = -1;
	}

	if (gp2x_B)
		gp2x_wait_for_key_release(gp2x_B);

	gp2x_unmute();
	return ret;
}

void gp2x_txt_msg(int x, int y, int c, const char *msg)
{
	textout_centre_ex(bmp, font, msg, x+1 , y+1, 32,-1);
	textout_centre_ex(bmp, font, msg, x ,   y,   c, -1);
}

void gp2x_display_vol()
{
	static int black = makecol( 0, 0, 0);
	int i;
	int volDisplay;
	char vol[50];

	int left   = 60;
	int right  = 271;
	int top    = 220;
	int bottom = top + 16;

	rectfill(bmp, left,   top,    right, bottom,  black);

	volDisplay = int(.25*app_data.svolume) - 1;

	for (i = 0; i < 25; i++)
	{
		if (i == volDisplay)
			vol[i] = '|';
		else
			vol[i] = '-';
	}

	vol[i] = '\0';
	
	gp2x_txt_msg(int(.5*(left+right+1)), top+4, 15+32, vol);
	finish_display();
	rest(100);
}

void gp2x_toggle_region()
{
	if (app_data.euro == 0)
		app_data.euro = 1;
	else
		app_data.euro = 0;

	//setvideomode(app_data.euro);
	gp2x_reset();
}

void gp2x_toggle_stick()
{
	int temp = app_data.stick[0];

	app_data.stick[0] = app_data.stick[1];
	app_data.stick[1] = temp;
}

void gp2x_std_stick()
{
	app_data.stick[0] = 2;
	app_data.stick[1] = 0;
}

void gp2x_swap_stick()
{
	app_data.stick[0] = 0;
	app_data.stick[1] = 2;
}

void gp2x_init_stick()
{
	gp2x_std_stick();

	//Satellite Attack
	if (app_data.crc == 0x395629E1) gp2x_swap_stick();
	if (app_data.crc == 0xB4735E6B) gp2x_swap_stick();
	if (app_data.crc == 0x2B1E9E19) gp2x_swap_stick();
	if (app_data.crc == 0x395629E1) gp2x_swap_stick();

	//Space Monster
	if (app_data.crc == 0x71AE8533) gp2x_swap_stick();

	//Race - Spin Out - Cryptogram
	if (app_data.crc == 0x4EDA4917) gp2x_swap_stick();
	if (app_data.crc == 0x3B24E1EE) gp2x_swap_stick();
	if (app_data.crc == 0xEE3EE642) gp2x_swap_stick();
	if (app_data.crc == 0x9BC04EBB) gp2x_swap_stick();

	//Cosmic Conflict
	if (app_data.crc == 0x2A5F0B2B) gp2x_swap_stick();
	if (app_data.crc == 0x2F6E4B46) gp2x_swap_stick();
	if (app_data.crc == 0xA8DBED92) gp2x_swap_stick();
	if (app_data.crc == 0xC049E79E) gp2x_swap_stick();
	if (app_data.crc == 0x105231F8) gp2x_swap_stick();
}

bool gp2x_choose_rom(char *rom)
{	
	int i;
	int nRoms;

	static int page      = 0;
	static int selection = 0;

	static int blue  = 9+32;
	static int white = 0+32;
	static int black = 15+32;
	int col          = white;

	int romsPerPage  = 15;
	int romsThisPage = 15;
	int height;
	int top;
	int textTop;
	int left;
	int right;
	int bottom;
	int nPage;
	int start;

	bool ret = false;

	gp2x_mute();

	gp2x_dir_list *files = gp2x_dirlist_load("./roms");
	nRoms = files->elem_count;

	if (romsPerPage > nRoms)
		romsPerPage = nRoms;

	height     = (14*romsPerPage);
	top        = 120 - (height/2);
	textTop    = top + 4;
	left       = 30;
	right      = 300;
	bottom     = textTop + height + 2;
	nPage      = 0;
	start      = 0;
	bottom     = textTop + height + 2;
	nPage      = int ((nRoms + romsPerPage - 1) / romsPerPage);

	do
	{
		rectfill(bmp, left,   top,    right, bottom,  blue);
		hline(   bmp, left,   top,    right,          15+32);
		hline(   bmp, left+1, bottom, right,          1+32);

		vline(   bmp, left,   top,    bottom,         15+32);
		vline(   bmp, right,  bottom, top,            1+32);

		start   = (page * romsPerPage);
		textTop = top + 4;

		romsThisPage = romsPerPage;
		if (page == nPage - 1)//last page
			romsThisPage = nRoms - (romsPerPage * (nPage-1));

		for (i = 0; i < romsThisPage; i++)
		{
			col = (selection == i ? white : black);

			gp2x_menu_txt(165,  textTop, col, files->elements[start + i].filename);
			gp2x_menu_txt(165,  textTop, col, files->elements[start + i].filename);
			textTop += 14;
		}

		finish_display();

		rest(100);
		gp2x_joystick_read();

		if (gp2x_up)
			selection--;

		if (gp2x_down)
			selection++;

		if (gp2x_L )
		{
			if (page > 0)
				page--;
			else
				selection = 0;
		}

		if (gp2x_R)
		{
			if (page < nPage - 1)
				page++;
			else
				selection = romsThisPage-1;
		}

		if (selection < 0)
		{
			if (page == 0)
				selection = 0;
			else 
			{
				selection = romsPerPage-1;
				page--;
			}
		}

		if (selection == romsThisPage)
		{
			if (page == nPage - 1)
				selection = romsThisPage - 1;
			else
			{
				selection = 0;
				page++;
			}
		}
	} 
	while(!(gp2x_B || gp2x_select));

	if (gp2x_select)
	{
		gp2x_wait_for_key_release(gp2x_select);
		ret = false;
	}

	if (gp2x_B)
	{
		gp2x_wait_for_key_release(gp2x_B);

		strcpy(rom,"roms/");
		strcat(rom, files->elements[selection+start].filename);

		ret = true;
	}

	gp2x_dirlist_destroy(files);
	gp2x_unmute();

	return ret;
}

bool gp2x_load_rom()
{
	char file[1024];

	if (!gp2x_choose_rom(file))
		return false;

	if (!gp2x_load_rom(file))
		return false;

	gp2x_reset();
	return true;
}

bool gp2x_load_rom(char *file)
{
	FILE *fn;
	long l;
	int i, nb;

	app_data.crc = crc32_file(file);

	if (app_data.crc == 0xAFB23F89) app_data.exrom = 1;  /* Musician */
	
	if (app_data.crc == 0x3BFEF56B) app_data.exrom = 1;  /* Four in 1 Row! */
	if (app_data.crc == 0x9B5E9356) app_data.exrom = 1;  /* Four in 1 Row! (french) */

	if ((app_data.crc == 0x4E2CC6D3) || (app_data.crc == 0xBCF1EFC9)) app_data.three_k = 1;  /* Kill the Attacking Aliens */

	init_system();
	gp2x_init_stick();
	
	fn=fopen(file,"rb");
	if (!fn) {
		fprintf(stderr,"Error loading %s\n",file);
		return false;
	}
	printf("Loading: \"%s\"  Size: ",file);
	l = gp2x_filesize(fn);

	if ((l % 1024) != 0) {
		fprintf(stderr,"Error: file %s is an invalid ROM dump\n",file);
		return false;
	}

	if (((l % 3072) == 0) && app_data.three_k) {

		nb = l/3072;

		for (i=nb-1; i>=0; i--) {
			if (fread(&rom_table[i][1024],3072,1,fn) != 1) {
				fprintf(stderr,"Error loading %s\n",file);
				return false;
			}
		}
		printf("%dK",nb*2);

	} else {

		nb = l/2048;

		if ((nb == 2) && (app_data.exrom)) {

			if (fread(&extROM[0], 1024,1,fn) != 1) {
				fprintf(stderr,"Error loading %s\n",file);
				return false;
			}
			if (fread(&rom_table[0][1024],3072,1,fn) != 1) {
				fprintf(stderr,"Error loading %s\n",file);
				return false;
			}
			printf("3K EXROM");

		} else {

			for (i=nb-1; i>=0; i--) {
				if (fread(&rom_table[i][1024],2048,1,fn) != 1) {
					fprintf(stderr,"Error loading %s\n",file);
					return false;
				}
				memcpy(&rom_table[i][3072],&rom_table[i][2048],1024); /* simulate missing A10 */
			}
			printf("%dK",nb*2);

		}
	}
	fclose(fn);

	rom = rom_table[0];

	if (nb==1)
			app_data.bank = 1;
	else if (nb==2)
		app_data.bank = app_data.exrom ? 1 : 2;
	else if (nb==4)
		app_data.bank = 3;
	else 
		app_data.bank = 4;

	if ((rom_table[nb-1][1024+12]=='O') && (rom_table[nb-1][1024+13]=='P') && (rom_table[nb-1][1024+14]=='N') && (rom_table[nb-1][1024+15]=='B')) app_data.openb=1;

	printf("  CRC: %08lX\n",app_data.crc);

	gp2x_set_state_name(file);

	return true;
}

void gp2x_set_state_name(char *file)
{
	int i;

	memset(gp2x_state, 0, 1024);

	strcpy(gp2x_state, "./states/");

	int len = strlen(file) - 3;
	for (i = 5; i < len; i++)
		gp2x_state[i+4] = file[i];

	strcat(gp2x_state, "sav");
}

long gp2x_filesize(FILE *stream)
{
   long curpos, length;

   curpos = ftell(stream);
   fseek(stream, 0L, SEEK_END);
   length = ftell(stream);
   fseek(stream, curpos, SEEK_SET);
   return length;
}

void gp2x_fill_entry(gp2x_dir_list *listing, gp2x_dir_ent *entry, const char *filename)
{
	struct stat st;
	char* path;
	
	path = ( char* ) malloc( strlen( listing->path ) + strlen( filename ) + 2 );

	strcpy( path, listing->path );
	strcat( path, "/" );
	strcat( path, filename );

	if ( stat( path, &st ) < 0 )
	{
		entry->type  = 0;
		entry->size  = 0;
		entry->perms = 0;
	}

	else
	{
		entry->perms = st.st_mode & ( S_IRWXU | S_IRWXG | S_IRWXO );

		if ( S_ISREG( st.st_mode ) )
			entry->type = '-';

		else if ( S_ISDIR( st.st_mode ) )
			entry->type = 'd';

		else if ( S_ISCHR( st.st_mode ) )
			entry->type = 'c';

		else if ( S_ISBLK( st.st_mode ) )
			entry->type = 'b';

		else if ( S_ISFIFO( st.st_mode ) )
			entry->type = 'f';

		else if ( S_ISLNK( st.st_mode ) )
			entry->type = '@';

		else if ( S_ISSOCK( st.st_mode ) )
			entry->type = 's';

		else
			entry->type = '?';

		entry->size  = st.st_size;
		entry->links = st.st_nlink;
	}

	free( path );
			
	entry->filename = strdup( filename );
}

void gp2x_dirlist_append(gp2x_dir_list *listing, const char* filename)
{
    if ( !strcmp( filename, "." ) )
        return;    /* skip it */

	if ( !strcmp( filename, ".." ) )
        return;    /* skip it */

	if ( listing->elem_count >= listing->elem_size )
	{
		gp2x_dir_ent *new_elem;
		
		listing->elem_size += GP2X_INCREASE_SIZE;
		new_elem = (gp2x_dir_ent *) malloc(sizeof(gp2x_dir_ent) *listing->elem_size);

		memcpy(new_elem, listing->elements, sizeof(gp2x_dir_ent) *listing->elem_count);
		free(listing->elements);

		listing->elements = new_elem;
	}

	gp2x_fill_entry(listing, listing->elements + listing->elem_count, filename);
	listing->elem_count++;
}

void gp2x_dirlist_destroy(gp2x_dir_list *listing)
{
	int i;
	
	for (i = 0; i < listing->elem_count; i++)
		free(listing->elements[i].filename);

	free(listing->elements);
	free(listing->path);
	free(listing);		
}

int gp2x_dir_ent_comp(const void* d1, const void* d2)
{
     gp2x_dir_ent *de1;
     gp2x_dir_ent *de2;

    de1 = (gp2x_dir_ent*)d1;
    de2 = (gp2x_dir_ent*)d2;

    if (de1->filename && de2->filename)
        return strcmp(de1->filename, de2->filename);

    return 0;
}

gp2x_dir_list *gp2x_dirlist_sort(gp2x_dir_list *listing)
{
    qsort(listing->elements, listing->elem_count, sizeof(gp2x_dir_ent), gp2x_dir_ent_comp);

    return listing;
}

gp2x_dir_list *gp2x_dirlist_load(const char* path)
{
	gp2x_dir_list *listing;
	DIR* dir;

	struct dirent *dirent;
	
	dir = opendir(path);

	if (dir == NULL)
		return NULL;
	
	listing             = (gp2x_dir_list *) malloc(sizeof(gp2x_dir_list));
	listing->path       = strdup(path);
	listing->elem_count = 0;
	listing->elem_size  = GP2X_INITIAL_SIZE;
	listing->elements   = (gp2x_dir_ent *) malloc(sizeof(gp2x_dir_ent) *GP2X_INITIAL_SIZE);

	while ((dirent = readdir(dir)) != NULL)
		gp2x_dirlist_append(listing, dirent->d_name);

	closedir(dir);

    listing = gp2x_dirlist_sort(listing);

	return listing;
}

char *gp2x_dirlist_get_path(gp2x_dir_list *listing, int n)
{
	char *path;
	
	path = (char *)malloc(strlen(listing->path) + strlen(listing->elements[n].filename) + 2);
	strcpy(path, listing->path);

	if (strcmp(path, "/") != 0)
		strcat(path, "/");

	strcat(path, listing->elements[n].filename);

	return path;
}

static int labelY;
static int labelX;

void gp2x_display_info()
{
	int height = 115;
	int top    = 120 - (height/2);
	int bottom = top + height;
	int left   = 20;
	int right  = 312;
	int y      = top;
	labelX     = 166;

	rectfill(bmp,  left,   top,    right, bottom,  9+32);
	line(    bmp,  left,   top,    right, top,    15+32);
	line(    bmp,  left,   top,    left,  bottom, 15+32);
	line(    bmp,  left+1, bottom, right, bottom,  1+32);
	line(    bmp,  right,  bottom, right, top,     1+32);

	y +=  4; gp2x_txt_msg(labelX, y, 15+32, "O2EM GP2X v1.05 Jan 2008");
	y += 14; gp2x_txt_msg(labelX, y, 15+32, "Free Odyssey2 / VP+ Emulator");
	y += 14; gp2x_txt_msg(labelX, y, 15+32, "Developed by Andre de la Rocha");
	y += 14; gp2x_txt_msg(labelX, y, 15+32, "and Arlindo M. de Oliveira");
	y += 14; gp2x_txt_msg(labelX, y, 15+32, "Copyright 1996/1998 by Daniel Boris");
	y += 24; gp2x_txt_msg(labelX, y, 15+32, "GP2X Port by Sean Poyser");

	labelY = y + 14;
}

void gp2x_display_loading()
{
	gp2x_display_info();
	gp2x_txt_msg(labelX, labelY, 15+32, "Loading...");

	finish_display();
}

void gp2x_display_loaded()
{
	gp2x_display_info();
	gp2x_txt_msg(labelX, labelY, 15+32, "Press B to continue");

	finish_display();

	do
	{
		gp2x_joystick_read();
	}
	while(!gp2x_B);

	gp2x_wait_for_key_release(gp2x_B);

	rectfill(bmp,  0,  0, 320, 240, 0);
	finish_display();
}

bool gp2x_load_bios(char *biosname)
{
	FILE *fn;
	char s[1024+10];
	char errTxt[100];
	unsigned long crc;
	int i;

	gp2x_bios = Unknown;
   
    if (!strcmp(biosname,"g7400"))
    {
		if ((!biosname) || (strlen(biosname)==0))
		{
			strcpy(s,"o2rom.bin");
			fn = fopen("o2rom.bin","rb");
			if (!fn) 
				fn = fopen("O2ROM.BIN","rb");
		} 
		else if ((biosname[strlen(biosname)-1]=='/') || (biosname[strlen(biosname)-1]=='\\') || (biosname[strlen(biosname)-1]==':')) 
		{
			strcpy(s,biosname);
			strcat(s,"O2ROM.BIN");
			fn = fopen(s,"rb");
			if (!fn) 
			{
				strcpy(s,biosname);
				strcat(s,"o2rom.bin");
				fn = fopen(s,"rb");
			}
		} 
		else 
		{
			strcpy(s,biosname);
			fn = fopen(biosname,"rb");
		}

		if (!fn)
		{
			sprintf(errTxt,"Error loading bios ROM (%s)", s);
			gp2x_display_msg_and_stop(errTxt);
			return false;
		}
	
 		if (fread(rom_table[0],1024,1,fn) != 1) 
		{
 			sprintf(errTxt, "Error loading bios ROM o2rom.bin");
			gp2x_display_msg_and_stop(errTxt);
 			return false;
 		}
    }
	    strcpy(s,biosname);
		fn = fopen(biosname,"rb");
	
    if (!fn) 
	{
		sprintf(errTxt, "Error loading bios ROM (%s)", s);
		gp2x_display_msg_and_stop(errTxt);
		return false;
	}
	
 	if (fread(rom_table[0],1024,1,fn) != 1) 
	{
 		sprintf(errTxt, "Error loading bios ROM o2rom.bin");
		gp2x_display_msg_and_stop(errTxt);
 		return false;
 	}
    
    fclose(fn);

	for (i = 1; i < 8; i++) 
		memcpy(rom_table[i],rom_table[0],1024);

	crc = crc32_buf(rom_table[0],1024);

	switch (crc)
	{
	case 0x8016A315:
		printf("Odyssey2 bios ROM loaded\n");
		app_data.vpp = 0;
		gp2x_bios    = G7000;
		break;

	case 0xE20A9F41:
		printf("Videopac+ G7400 bios ROM loaded\n");
		app_data.vpp = 1;
		gp2x_bios    = G7400;
		break;

	case 0xA318E8D6:
		printf("C52 bios ROM loaded\n");
		app_data.vpp = 0;
		gp2x_bios    = C52;
		break;

	case 0x11647CA5:
		printf("Jopac bios ROM loaded\n");
		app_data.vpp = 1;
		gp2x_bios    = JOPAC;
		break;

	default:
		printf("Bios ROM loaded (unknown version)\n");
		app_data.vpp = 0;
		gp2x_bios    = Unknown;
		break;
	}

	return true;
}

int gp2x_savestate()
{
	return savestate(gp2x_state);
}

int gp2x_loadstate()
{
	remove("/mnt/minimal.o");
	//if (gp2x_bios == G7000)
		return loadstate(gp2x_state);

	//kludge to load state in G7400 bios
	//will only work if '0' can be used to start game
	//leave out, will look into later
	//gp2x_reset();
	//gp2x_flush_keyboard();
	//key[KEY_0]      = true;
	//gp2x_load_state = true;

	return 0;
}

void gp2x_display_msg(char *msg, int waits)
{
	gp2x_mute();

	rectfill(bmp, 0, 0, 320, 240, 0);

	gp2x_txt_msg(166, 120, 15+32, msg);

	finish_display();

	rest(waits*100);

	gp2x_unmute();
}

void gp2x_display_msg_and_stop(char *msg)
{
	gp2x_mute();

	rectfill(bmp, 0, 0, 320, 240, 0);

	gp2x_txt_msg(166, 120, 15+32, msg);

	finish_display();

	do
	{
		gp2x_joystick_read();
	}
	while(!gp2x_B);

	gp2x_wait_for_key_release(gp2x_B);

	gp2x_unmute();
}
